/***************************************************************************
                          clientrender.cpp  -  description
                             -------------------
    begin                : Thu Feb 27 2003
    copyright            : (C) 2003 by Silviu Simen
    email                : ssimen@vryad.com
 ***************************************************************************/
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <math.h>
#include <stdio.h>
#include <time.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "clientdefines.h"
#include "p5sdk.h"
#include "p5client.h"
#include "estimator.h"

#define POS_FILTER_SIZE 10
#define ROT_FILTER_SIZE 17

P5Data p5data;

t_p5_vec3 cen, ypr, gpos, pos, zeropos, zerorot;
float glove_error;
float gl_rescale = 1.0 / 100.0f;
int use_abs_pos = 1;
int finger_bend[5];

CEstimator est_pos[3];
CEstimator est_ypr[3];

void ProcessGloveInput()
{
  int k;
  for (k=0;k<3;k++)
  {
    est_pos[k].AddSample(gpos.v[k]);
    est_ypr[k].AddSample(ypr.v[k]);
  };

  pos.x = -(est_pos[0].GetEstimate() - zeropos.x) * gl_rescale;
  pos.y = +(est_pos[1].GetEstimate() - zeropos.y) * gl_rescale;
  pos.z = +(est_pos[2].GetEstimate() - zeropos.z) * gl_rescale;

  ypr.yaw = est_ypr[0].GetEstimate() - zerorot.yaw;
  ypr.pitch = est_ypr[1].GetEstimate() - zerorot.pitch;
  ypr.roll = -(est_ypr[2].GetEstimate() - zerorot.roll);

}

void GetGloveInput()
{
  cli_GetP5Data (0, &p5data);
  glove_error = p5data.m_error_estimate;
  if (use_abs_pos)
  {
    gpos = p5data.m_absolute_pos;
    ypr = p5data.m_absolute_rot;
  }
  else
  {
    gpos = p5data.m_relative_pos;
    ypr = p5data.m_relative_rot;
  }

  for (int i=0;i<5;i++)
     finger_bend[i] = p5data.m_bend[i];

  ProcessGloveInput();
}

void InitPRFilters ()
{
  int k;
  for (k=0;k<3;k++)
  {
    est_pos[k].ZeroAll();
    est_ypr[k].ZeroAll();
  };
}

void GetZeroPosition ()
{
  if (use_abs_pos)
  {
    zeropos = p5data.m_absolute_pos;
    zerorot.yaw = 0.0f;
    zerorot.pitch = 0.0f;
    zerorot.roll = 0.0f;
  }
  else
  {
    zeropos = p5data.m_relative_pos;
    zerorot = p5data.m_relative_rot;
  }
}

void EngineInit ()
{
  gl_rescale = (SCREENX > SCREENY) ? 5.0 / (float) SCREENY : 5.0 / (float) SCREENX;
  srand (time (NULL));
  est_pos[0].Init (POS_FILTER_SIZE);
  est_pos[1].Init (POS_FILTER_SIZE);
  est_pos[2].Init (POS_FILTER_SIZE);
  est_ypr[0].Init (ROT_FILTER_SIZE);
  est_ypr[1].Init (ROT_FILTER_SIZE);
  est_ypr[2].Init (ROT_FILTER_SIZE);

  GetGloveInput();
  GetZeroPosition();
}

void EngineDone ()
{
  static int cleanedup = 0;

  if (!cleanedup)
  {
    cleanedup = 1;
  }
}

void Create3DHand(float size=1.0f)
{
	const int NPOINTS=8;
	const float pscale=0.1f;
	const float zsize=0.4f;
	int i;
	GLfloat pdata[NPOINTS][2] = 
	{
		{20,63}, {16,47}, {25,29}, {35,28}, {43,31}, {51,36}, {52,49}, {45,63}
	};

	GLfloat fingers_base[5][2] = 
	{
		{17,53}, {25,28}, {35,28}, {43,30}, {51,37}
	};
	GLfloat fingers_top[5][2] = 
	{
		{1,36}, {25,5}, {35,0}, {43,5}, {51,21}
	};
	
	// hand color
	glColor3f( 251.0f/355.0f, 156.0f/355.0f, 112.0f/355.0f);

	float xc=0.0f,yc=0.0f;
	for (i=0;i<NPOINTS;i++)
	{
		xc += pscale*pdata[i][0];
		yc += pscale*pdata[i][1];
	};
	xc /= (float)NPOINTS;
	yc /= (float)NPOINTS;

	glTranslatef(-xc, -yc, -zsize);

	glBegin(GL_QUADS);
	glVertex3f(pscale*pdata[NPOINTS-1][0],pscale*pdata[NPOINTS-1][1], zsize);
	glVertex3f(pscale*pdata[NPOINTS-1][0],pscale*pdata[NPOINTS-1][1], -zsize);
	glVertex3f(pscale*pdata[0][0],pscale*pdata[0][1], -zsize);
	glVertex3f(pscale*pdata[0][0],pscale*pdata[0][1], zsize);
	glEnd();

	
	for (i=0;i<NPOINTS-1;i++)
	{
		glBegin(GL_QUADS);
		glVertex3f(pscale*pdata[i][0],pscale*pdata[i][1], zsize);
		glVertex3f(pscale*pdata[i][0],pscale*pdata[i][1], -zsize);
		glVertex3f(pscale*pdata[i+1][0],pscale*pdata[i+1][1], -zsize);
		glVertex3f(pscale*pdata[i+1][0],pscale*pdata[i+1][1], zsize);
		glEnd();
	}

	glBegin(GL_POLYGON);
	for (i=0;i<NPOINTS;i++)
	{
		glVertex3f(pscale*pdata[i][0],pscale*pdata[i][1], zsize);
	}
	glEnd();

	glBegin(GL_POLYGON);
	for (i=NPOINTS-1;i>=0;i--)
	{
		glVertex3f(pscale*pdata[i][0],pscale*pdata[i][1], -zsize);
	}
	glEnd();

	// finger colors
	glColor3f( 251.0f/255.0f, 156.0f/255.0f, 112.0f/255.0f);

	for (i=0;i<5;i++)
	{
		#define IS_THUMB (i==0)
		GLUquadricObj *quadObj;
		quadObj = gluNewQuadric ();
		gluQuadricDrawStyle (quadObj, GLU_FILL);
		gluQuadricNormals (quadObj, GLU_SMOOTH);

		float xf = 90 + atan2(
				fingers_top[i][1] - fingers_base[i][1], 
				fingers_top[i][0] - fingers_base[i][0]) * 180.0f / 3.14159f;

		float tsize = 1.0 * pscale * sqrt(
				pow(fingers_top[i][1] - fingers_base[i][1],2) + 
				pow(fingers_top[i][0] - fingers_base[i][0],2) );


		glTranslatef(pscale*fingers_base[i][0], pscale*fingers_base[i][1], 0);

		//glutSolidTorus(1,5, 10,20);

		int nsf = 8;
		int numslices = 5;
		int numstacks = 5;
		int n;
		float radius = zsize*0.8;
		
		for (n=0; n<nsf; n++)
		{
			float rangle;
			if (IS_THUMB)
				rangle = -45 + 3 * n * finger_bend[i]/(float)nsf;
			else
				rangle = 90 - 3 * n * finger_bend[i]/(float)nsf;

			if (IS_THUMB) // thumb rotates differently
				glRotatef( +rangle, 0, 1, 0);
			else
				glRotatef( +rangle, 1, 0, 0);
			glRotatef( xf, 0, 1, 0);
			
			float partsize = tsize/(float)nsf;

			glTranslatef(0,0, n*partsize+partsize/2);

			if (IS_THUMB)
				glRotatef( +3 * finger_bend[i]/(float)nsf, 0, 1, 0);
			else
				glRotatef( -3 * finger_bend[i]/(float)nsf, 1, 0, 0);

			gluCylinder(quadObj, radius, radius, partsize/2, numslices, numstacks);

		        glTranslatef(0,0, partsize/2);
		        gluSphere(quadObj, radius, numslices, numstacks);
		        glTranslatef(0,0, -partsize/2);
			
			glRotatef( 180, 0, 1, 0);

			gluCylinder(quadObj, radius, radius, partsize/2, numslices, numstacks);

		        glTranslatef(0,0, partsize/2);
		        gluSphere(quadObj, radius, numslices, 10);
		        glTranslatef(0,0, -partsize/2);

			glRotatef( -180, 0, 1, 0);

			if (IS_THUMB)
				glRotatef( -3 * finger_bend[i]/(float)nsf, 0, 1, 0);
			else
				glRotatef( +3 * finger_bend[i]/(float)nsf, 1, 0, 0);
			
			glTranslatef(0,0, -n*partsize-partsize/2);

			glRotatef( -xf, 0, 1, 0);
			if (IS_THUMB) // thumb rotates differently
				glRotatef( -rangle, 0, 1, 0);
			else
				glRotatef( -rangle, 1, 0, 0);
		}


	        glTranslatef(-pscale*fingers_base[i][0], -pscale*fingers_base[i][1], 0);
	        gluDeleteQuadric(quadObj);
	}


/*	
	for (i=0;i<5;i++)
	{
		#define IS_THUMB (i==0)
		GLUquadricObj *quadObj;
		quadObj = gluNewQuadric ();
		gluQuadricDrawStyle (quadObj, GLU_FILL);
		gluQuadricNormals (quadObj, GLU_SMOOTH);

		float xf = 90 + atan2(
				fingers_top[i][1] - fingers_base[i][1], 
				fingers_top[i][0] - fingers_base[i][0]) * 180.0f / 3.14159f;

		float tsize = pscale * sqrt(
				pow(fingers_top[i][1] - fingers_base[i][1],2) + 
				pow(fingers_top[i][0] - fingers_base[i][0],2) );

		float rxangle;
		if (IS_THUMB)
			rxangle = -45 + 3 * finger_bend[i];
		else
			rxangle = 90 - 2.5 * finger_bend[i];

		glTranslatef(pscale*fingers_base[i][0], pscale*fingers_base[i][1], 0);
		if (IS_THUMB) // thumb rotates differently
			glRotatef( +rxangle, 0, 1, 0);
		else
			glRotatef( +rxangle, 1, 0, 0);
		glRotatef( xf, 0, 1, 0);
		
		gluCylinder(quadObj, zsize*0.8, zsize*0.4, tsize, 15, 10);
	        gluSphere(quadObj, zsize*0.8, 15, 10);
	        
	        glTranslatef(0,0, tsize);
	        gluSphere(quadObj, zsize*0.4, 15, 10);
	        glTranslatef(0,0, -tsize);

		glRotatef( -xf, 0, 1, 0);
		if (IS_THUMB) // thumb rotates differently
			glRotatef( -rxangle, 0, 1, 0);
		else
			glRotatef( -rxangle, 1, 0, 0);

	        glTranslatef(-pscale*fingers_base[i][0], -pscale*fingers_base[i][1], 0);
	        gluDeleteQuadric(quadObj);
	}
*/
	glTranslatef(xc, yc, zsize);
}


void Put3DData ()
{
  glTranslatef (pos.x, pos.y, pos.z);

  glRotatef( 135, 1.0f, 0.0f, 0.0f);
  
  glRotatef (ypr.roll, 0.0f, 1.0f, 0.0f);
  glRotatef (ypr.pitch, 1.0f, 0.0f, 0.0f);
  glRotatef (ypr.yaw, 0.0f, 0.0f, 1.0f);

  Create3DHand();
}

void SetParameters ()
{
  glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // setup texture mapping
  glEnable (GL_TEXTURE_2D);
  glEnable (GL_DEPTH_TEST);
  glEnable (GL_CULL_FACE);
  glShadeModel (GL_SMOOTH);

  glCullFace (GL_BACK);
}

void SetCamera ()
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  gluPerspective (30, SCREENX / SCREENY, 0.1, 2000);
  glTranslatef (0, 0, -40.f);
  glViewport (0, 0, SCREENX, SCREENY);
}

void ErrorBar (float error)
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glOrtho (0, SCREENX, 0, SCREENY, -1.0, 1.0);


  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  glDisable (GL_TEXTURE_2D);
  glColor3f (1.0f, 1.0f, 0.0f);

  glShadeModel (GL_FLAT);

  int x_size = (int) (error * SCREENX / 100.0f);
  int y_size = 20;

  GLint squareShapeVertices[] = {
    0, SCREENY - 0, 0,
    0, SCREENY - y_size, 0,
    x_size, SCREENY - y_size, 0,
    x_size, SCREENY - 0, 0
  };
  GLubyte squareVertexColours[] = {
    255, 25, 255,
    255, 25, 255,
    255, 25, 255,
    255, 25, 255
  };

  glEnableClientState (GL_VERTEX_ARRAY);
  glEnableClientState (GL_COLOR_ARRAY);
  glVertexPointer (3, GL_INT, 0, squareShapeVertices);
  glColorPointer (3, GL_UNSIGNED_BYTE, 0, squareVertexColours);
  glDrawArrays (GL_QUADS, 0, 4);

  int i;
  glBegin (GL_QUADS);
  for (i = 0; i < 4; i++)
    glArrayElement (i);
  glEnd ();

}

void ConsoleLine (const char *message)
{
  static int crtline = 0;

  if (message == NULL)
  {
    crtline = 0;
    return;
  }


  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glOrtho (0, SCREENX, 0, SCREENY, -1.0, 1.0);


  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  int posy = SCREENY - 70;
  int posx = 10;
  int deltay = 15;


  glDisable (GL_TEXTURE_2D);
  glColor3f (1.0f, 1.0f, 0.0f);

  glShadeModel (GL_FLAT);
  glRasterPos2i (posx, posy - crtline * deltay);
  
  for (char c=0; message[c] != 0; c++) 
  	glutBitmapCharacter(GLUT_BITMAP_8_BY_13, message[c]);

  crtline++;
}

void PutFrameRate ()
{
  char buffer[1000];
  ConsoleLine (0);

  ConsoleLine ("Info : ");

  sprintf (buffer, "xobj=%4.1f", pos.x);
  ConsoleLine (buffer);
  sprintf (buffer, "yobj=%4.1f", pos.y);
  ConsoleLine (buffer);
  sprintf (buffer, "zobj=%4.1f", pos.z);
  ConsoleLine (buffer);

  sprintf (buffer, "yaw=%4.1f", ypr.yaw);
  ConsoleLine (buffer);
  sprintf (buffer, "pitch=%4.1f", ypr.pitch);
  ConsoleLine (buffer);
  sprintf (buffer, "roll=%4.1f", ypr.roll);
  ConsoleLine (buffer);

  sprintf (buffer, "error estimate=%4.1f", glove_error);
  ConsoleLine (buffer);
}

void Render ()
{
  GetGloveInput ();
  SetParameters ();
  SetCamera ();

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  glDisable (GL_TEXTURE_2D);
  glColor3f (1.0f, 1.0f, 1.0f);

  glutWireCube ((float) SCREENX * gl_rescale * 2.0f);
  glTranslatef (cen.x, cen.y, -cen.z);

  Put3DData ();
  ErrorBar (glove_error);
  PutFrameRate ();
  glFlush ();


  glutSwapBuffers ();
  glutPostRedisplay ();

  return;
}

void myKeyboard (unsigned char k, int mouseX, int mouseY)
{
  switch (k)
  {
  case 27:
    exit (0);
    break;

  case 'a':
    use_abs_pos = 1;
    GetZeroPosition ();
    InitPRFilters();
    GetGloveInput();
    break;

  case 'r':
    use_abs_pos = 0;
    GetZeroPosition ();
    InitPRFilters();
    GetGloveInput();
    break;

  default:
    return;
  }
}
