/***************************************************************************
                          p5mouse.cpp  -  description
                             -------------------
    begin                : Tue Jan 28 2003
    copyright            : (C) 2003 by Silviu Simen
    email                : ssimen@vryad.com
 ***************************************************************************/

#include "p5sdk.h"
#include "p5server.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>

#define BUTTONS 3

#define BUTTON1 0x04
#define BUTTON2 0x02
#define BUTTON3 0x01

#define MAP_BUTTON1 0
#define MAP_BUTTON2 1
#define MAP_BUTTON3 2

CP5MouseRepeater::CP5MouseRepeater()
{
    filter_size = 1;
}

void CP5MouseRepeater::mous_CreateFilter(int fsize)
{
  filter_size = fsize;
  filterx = (float*) calloc(fsize,sizeof(float));
  filtery = (float*) calloc(fsize,sizeof(float));
  filterz = (float*) calloc(fsize,sizeof(float));
}

void CP5MouseRepeater::mous_RemoveFilter()
{
  free (filterx);
  free (filtery);
  free (filterz);
}

void CP5MouseRepeater::mous_FilterAdd(float x, float y, float z)
{
  int i;
  for (i = filter_size-1; i > 0; i--)
  {
    filterx[i] = filterx[i-1];
    filtery[i] = filtery[i-1];
    filterz[i] = filterz[i-1];
  }
  filterx[0] = x;
  filtery[0] = y;
  filterz[0] = z;
}

void CP5MouseRepeater::mous_FilterGet(float *x, float *y, float *z)
{
  int i;
  *x=0;
  *y=0;
  *z=0;
  for (i = filter_size; i >= 0; i--)
  {
    *x += filterx[i];
    *y += filtery[i];
    *z += filterz[i];
  }
  *x /= (float) filter_size;
  *y /= (float) filter_size;
  *z /= (float) filter_size;
}

void CP5MouseRepeater::mous_Setup(int mouse_type)
{
  m_type = mouse_type;
  
  char * device_file="";
  if (m_type == type_msc)
    device_file = MSC_FILE;
  if (m_type == type_imps2)
    device_file = IMPS_FILE;
  filed=open(device_file,O_RDWR|O_NONBLOCK);
  if(filed==-1)
  {
    int errorcode = mkfifo(device_file,0644);
    if (errorcode == 0)
    {
	filed=open(device_file,O_RDWR|O_NONBLOCK);
	if (filed == -1)
	{
	    printf("Could not open fifo %s: \n",device_file);
	    exit(0);
	}
    }
    else
    {
	printf("Could not create fifo %s: \n",device_file);
	exit(0);
    }
  }
  mous_CreateFilter(3);
}

void CP5MouseRepeater::mous_Close()
{
  //printf("P5 Mouse got signal, cleaning up !\n");
  shutdown(filed,2);
  close(filed);
  mous_RemoveFilter();
}


void CP5MouseRepeater::mous_msc_RawSend (int dx, int dy, int dz, int buttp, int buttr)
{
  static int buttons = 0;
  char buffer[5];

  buttons |= buttp;
  buttons &= ~buttr;

  buffer[0] = ~(buttons | 0x78);

  buffer[1] = dx;
  buffer[2] = dy;
  buffer[3] = buffer[4] = 0;

//  printf("dx=%d dy=%d fileid=%d \n", dx,dy, filed);

  write (filed, buffer, 5);
}

void CP5MouseRepeater::mous_msc_Move (int dx, int dy, int dz)
{
  mous_msc_RawSend (dx, dy, dz, 0, 0);
}

void CP5MouseRepeater::mous_msc_Button (int down, int up)
{
  mous_msc_RawSend (0, 0, 0, down, up);
}

//P5Data last_p5data;
void CP5MouseRepeater::mous_msc_ProcessData(P5Data *p5data)
{
  float lx=0, ly=0, lz=0;
  float cx=0, cy=0, cz=0;
//  printf("x=%f y=%f \n", p5data->m_relative_pos.v[0], p5data->m_relative_pos.v[1]);

  mous_FilterGet(&lx, &ly, &lz);
  mous_FilterAdd(p5data->m_relative_pos.v[0], p5data->m_relative_pos.v[1],0);
  mous_FilterGet(&cx, &cy, &cz);

  int dx = (int)-(cx - lx);
  int dy = (int)+(cy - ly);

//  printf("x=%f y=%f  dx=%d dy=%d\n", p5data->m_relative_pos.v[0], p5data->m_relative_pos.v[1], dx,dy);

  mous_msc_Move (dx , dy, 0);
  int down,up;
  down = ( (p5data->m_bend[1] > 20) ? 4 : 0 ) +
         ( (p5data->m_bend[0] > 20) ? 2 : 0 ) +
         ( (p5data->m_bend[2] > 20) ? 1 : 0 ) ;
  up   = ( (p5data->m_bend[1] < 20) ? 4 : 0 ) +
         ( (p5data->m_bend[0] < 20) ? 2 : 0 ) +
         ( (p5data->m_bend[2] < 20) ? 1 : 0 ) ;
  mous_msc_Button (down, up);
}

void CP5MouseRepeater::mous_ProcessData(P5Data *p5data)
{
  if (m_type == type_msc)
    mous_msc_ProcessData(p5data);
    
  if (m_type == type_imps2)
    mous_imps2_ProcessData(p5data);
}

void CP5MouseRepeater::mous_imps2_ProcessData(P5Data *p5data)
{
  float lx=0, ly=0, lz=0;
  float cx=0, cy=0, cz=0;

  mous_FilterGet(&lx, &ly, &lz);
  mous_FilterAdd(p5data->m_relative_pos.v[0], p5data->m_relative_pos.v[1], p5data->m_relative_pos.v[2]);
  mous_FilterGet(&cx, &cy, &cz);

  int dx = (int)-(cx - lx) / 2;
  int dy = (int)-(cy - ly) / 2;
  int dz = (int)+(cz - lz);

  int b[5];
  for (int i=0;i<5;i++)
    b[i] = p5data->m_bend[i]>20 ? 1 : 0 ;
  char buffer[4];

  if (dx > 255) dx = 255;
  if (dx < -256) dx = -256;
  if (dy > 255) dy = 255;
  if (dy < -256) dy = -256;
   
  buffer[0] = 8 |(b[1] ? 1 : 0) | (b[0] ? 4 : 0) | (b[2] ? 2 : 0) | (dx < 0 ? 0x10 : 0) |  (dy > 0 ? 0x20 : 0);
  buffer[1] = dx & 0xFF;
  buffer[2] = (-dy) & 0xFF;
  if (b[4])
     buffer[3] = (dz > 30 ? -1 : 0) + (dz < -30 ? 1 : 0);
  else
     buffer[3] = 0;
  //printf("B0=%.3u   dx = %+.3d, dy=%+.3d, dz=%+.3d   PINKY=%d\n", buffer[0], dx,dy,dz, b[4]);
  write (filed, buffer, 4);
}
