/***************************************************************************
                          p5server.cpp  -  description
                             -------------------
    begin                : Thu Feb 27 2003
    copyright            : (C) 2003 by Silviu Simen
    email                : ssimen@vryad.com
 ***************************************************************************/

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>

#include "p5sdk.h"

int reuse_mem = 0;
int usb_debuglevel = 0;
int go_daemon = 0;
int export_mouse = 0;
int export_imps2_mouse = 0;
int helpmode = 0;
int versionmode = 0;
int calibratemode = 0;
int calibrate_gloveno = 0;
int rotation_mode = -1;

int sharedmemid;
int semaphoreid;

extern P5SharedMem *shm_data;

void srv_SetupOptions (int argc, char **argv)
{
  int opt;
  opterr = 0;

  while ((opt = getopt (argc, argv, "brdmhvic:o:")) != -1)
  {
    switch (opt)
    {
    case 'b':
      go_daemon = 1;
      break;
    case 'r':
      reuse_mem = 1;
      break;
    case 'd':
      usb_debuglevel = 2;
      break;
    case 'm':
      export_mouse = 1;
      break;
    case 'i':
      export_imps2_mouse = 1;
      break;
    case 'h':
      helpmode = 1;
      break;
    case 'v':
      versionmode = 1;
      break;
    case 'c':
      calibratemode = 1;
      if (optarg == NULL)
      	calibrate_gloveno = 0;
      else
      	calibrate_gloveno = atoi(optarg);
      break;
    case 'o':
      rotation_mode = -1;
      if (optarg != NULL)
      	rotation_mode = atoi(optarg);
      break;
    case '?':
      if (optopt == 'c')
      {
      	calibratemode = 1;
      	calibrate_gloveno = 0;
      }
     break;
    };
  }
}

void srv_Close ()
{
  shmdt (shm_data);
  if (!reuse_mem)
  {
    if (sharedmemid != -1)
      shmctl (sharedmemid, IPC_RMID, NULL);
    if (semaphoreid != -1)
      semctl (semaphoreid, 0, IPC_RMID);
  }
}

void main_Cleanup(); // in server main

void srv_SigHandler (int)
{
  main_Cleanup();
  exit (0);
}

void srv_CatchSignals ()
{
  signal (SIGTERM, srv_SigHandler);
  signal (SIGQUIT, srv_SigHandler);
  signal (SIGINT, srv_SigHandler);
}

void srv_SetOptions (int argc, char **argv)
{
  srv_SetupOptions (argc, argv);

  if (go_daemon)
    daemon (1, 1);              // go in background ...
}

void sem_P (int semid)
{
  struct sembuf psembuf;
  psembuf.sem_num = 0;
  psembuf.sem_op = -1;
  psembuf.sem_flg = 0;
  semop (semid, &psembuf, 1);
}

void sem_V (int semid)
{
  struct sembuf vsembuf;
  vsembuf.sem_num = 0;
  vsembuf.sem_op = 1;
  vsembuf.sem_flg = 0;
  semop (semid, &vsembuf, 1);
}

void srv_Init (int num_p5)
{
  int errorcode = mkfifo(SHM_FILE,0644);
  if (errorcode == -1)
  {
    if (errno != EEXIST)
    {
	printf ("There was an error creating the fifo file [%s]!\n",SHM_FILE);
	exit (0);
    }
  }
  key_t p5key = ftok (SHM_FILE, SHM_PROJ);
  if (p5key == -1)
  {
    printf ("Cannot create shared memory key !\n");
    exit (0);
  }

  int sharedsize = (num_p5 - 1) * sizeof (P5Data) + sizeof (P5SharedMem);

  sharedmemid = shmget (p5key, sharedsize, IPC_CREAT | IPC_EXCL | 0666);
  if (sharedmemid == -1)
  {
    if (errno == EEXIST)
    {
      if (reuse_mem)
      {
        sharedmemid = shmget (p5key, sharedsize, 0666);
        if (sharedmemid == -1)
        {
          printf ("Cannot get shared memory !\n");
          exit (0);
        }
      }
      else
      {
        printf ("Shared memory block already exists (key=0x%08x size=%d) !\n", p5key, sharedsize);
        printf ("This might happened if server crashed or some other process is using that memory block.\n");
        printf ("You can remove the memory block this way : \n");
        printf ("  ipcrm -m `ipcs -m | grep 0x%08x | awk '{print $2}'` \n", p5key);
        exit (0);
      }
    }
    else
    {
      printf ("Cannot create shared memory !\n");
      exit (0);
    }
  }

  shm_data = (P5SharedMem *) shmat (sharedmemid, NULL, 0);
  if ((int) shm_data == -1)
  {
    printf ("Error attaching shared memory block !\n");
    exit (0);
  }

  semaphoreid = semget (p5key, 1, IPC_CREAT | IPC_EXCL | 0666);
  if (semaphoreid == -1)
  {
    if (errno == EEXIST)
    {
      if (reuse_mem)
      {
        semaphoreid = semget (p5key, 1, 0666);
        if (semaphoreid == -1)
        {
          srv_Close ();
          printf ("Cannot get semaphore !\n");
          exit (0);
        }
      }
      else
      {
        printf ("Semaphore already exists (key=0x%08x) !\n", p5key);
        printf ("This might happened if server crashed or some other process is using that semaphore.\n");
        printf ("You can remove the semaphore this way : \n");
        printf ("  ipcrm -s `ipcs -s | grep 0x%08x | awk '{print $2}'` \n", p5key);
        srv_Close ();
        exit (0);
      }
    }
    else
    {
      printf ("Error creating semaphore !\n");
      srv_Close ();
      exit (0);
    }
  }
  semctl (semaphoreid, 0, SETVAL, 1);

  sem_P (semaphoreid);
  shm_data->nump5s = num_p5;
  shm_data->p5datasize = sizeof (P5Data);
  sem_V (semaphoreid);

  srv_CatchSignals ();
}

void srv_CopyData (int j, P5Data * p5data)
{
  sem_P (semaphoreid);
  shm_data->p[j] = *p5data;
  sem_V (semaphoreid);
}
