/***************************************************************************
                          servermain.cpp  -  description
                             -------------------
    begin                : Tue Jan 28 2003
    copyright            : (C) 2003 by Silviu Simen
    email                : ssimen@vryad.com
 ***************************************************************************/

#include "p5sdk.h"
#include "p5server.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

int i;
int num_p5 = 0;

void PrintP5Data (P5Data * p)
{
  if (0)
    printf ("%1d %1d %1d %1d   ", p->m_buttons[0], p->m_buttons[1], p->m_buttons[2], p->m_buttons[3]);
  if (0)
    printf ("%3d %3d %3d %3d %3d   ", p->m_bend[0], p->m_bend[1], p->m_bend[2], p->m_bend[3], p->m_bend[4]);
  if (0)
    printf ("%9.2f %9.2f %9.2f   ", p->m_absolute_pos.v[0], p->m_absolute_pos.v[1], p->m_absolute_pos.v[2]);
  if (0)
    printf ("%9.2f %9.2f %9.2f   ", p->m_relative_pos.v[0], p->m_relative_pos.v[1], p->m_relative_pos.v[2]);
  if (0)
    printf ("%9.2f %9.2f %9.2f   ", p->m_absolute_rot.v[0], p->m_absolute_rot.v[1], p->m_absolute_rot.v[2]);
  if (0)
    printf ("%9.2f %9.2f %9.2f   ", p->m_relative_rot.v[0], p->m_relative_rot.v[1], p->m_relative_rot.v[2]);
  if (0)   { printf ("\n"); fflush (stdout); };
}

P5SharedMem *shm_data;
CP5MouseRepeater mice[2];

void main_Cleanup()
{
  srv_Close();
  mice[0].mous_Close();
  mice[1].mous_Close();
}
extern int export_mouse;
extern int export_imps2_mouse;
extern int helpmode;
extern int versionmode;
extern int calibratemode;
extern int calibrate_gloveno;
extern int rotation_mode;

int glove_Calibrate(int calibrate_gloveno)
{
  char buffer[100];
  printf("Calibrating P5 glove %d ... \n", calibrate_gloveno);
  
  printf("Unbend your fingers to a comfortable position. \n");
  printf("Press ENTER to continue : "); fgets(buffer, 99, stdin);
  
  drv_CalibrateDeviceStart(calibrate_gloveno);
  printf("Make a fist, bending all fingers; then unbend to a normal position.\n");
  
  printf("Press ENTER to continue : "); fgets(buffer, 99, stdin);
  drv_CalibrateDeviceStop(calibrate_gloveno);

  return 0;
}

int main (int argc, char **argv)
{
  srv_SetOptions (argc, argv);
  if (helpmode)
  {
    printf("Usage  : %s [-h] [-b] [-r] [-d] [-m] [-i] [-c glove_no] [-o mode ]\n", argv[0]);
    printf(" where : \n");
    printf("          -h   =  help (you see it now)  \n");
    printf("          -b   =  go background (daemon) \n");
    printf("          -r   =  reuse shared memory/semaphore \n");
    printf("          -d   =  USB debug mode (see libusb) \n");
    printf("          -m   =  export mouse data to %s \n", MSC_FILE);
    printf("          -i   =  export mouse data to %s \n", IMPS_FILE);
    printf("          -v   =  print version \n");
    printf("          -c   =  calibrate glove glove_no \n");
    printf("          -o   =  set rotation calculating mode \n");
    exit(0);
  };

  if (versionmode)
  {
    printf("%s version %s built on %s at %s\n", argv[0], P5_DRIVER_VERSION, __DATE__, __TIME__);
    printf(" %s -h for help \n", argv[0]);
    exit(0);
  };
  
  num_p5 = drv_InitP5 ();
  if (num_p5 <= 0)
  {
    exit(0);
  };

  if (calibratemode)
  {
    if (calibrate_gloveno < num_p5)
    	glove_Calibrate(calibrate_gloveno);
    else
    	printf("P5 Glove [%d] not detected !\n",calibrate_gloveno);
    exit(0);
  };

  srv_Init (num_p5);
  mice[0].mous_Setup(CP5MouseRepeater::type_msc);
  mice[1].mous_Setup(CP5MouseRepeater::type_imps2);
  P5Data p5data[num_p5];
  
  if (rotation_mode != -1)
  {
  	printf("Setting rotation calculating mode to [%d].\n",rotation_mode);
  	drv_SetRotationCalculationMode (rotation_mode);
  }

  while (1)
  {
    for (int j = 0; j < num_p5; j++)
    {
      drv_ProcessOneFrame (j, &(p5data[j]));
      srv_CopyData (j, &(p5data[j]));
      
      PrintP5Data (&(p5data[j]));

      // we support for now only one P5 in mouse mode
      // see p5mouse.cpp ...
      if (  (j==0) && (export_mouse) )
      {
        mice[0].mous_ProcessData(&(p5data[j]));
      };
      if (  (j==0) && (export_imps2_mouse) )
      {
        mice[1].mous_ProcessData(&(p5data[j]));
      };
    }
  }
  main_Cleanup();
}
