/***************************************************************************
                          starfield.cpp  -  description
                             -------------------
    begin                : Thu May 05 2003
    copyright            : (C) 2003 by Silviu Simen
    email                : ssimen@vryad.com
 ***************************************************************************/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <GL/glut.h>

#include "p5sdk.h"
#include "p5client.h"
#include "estimator.h"

#define POS_FILTER_SIZE 5
#define ROT_FILTER_SIZE 7
P5Data p5data;
CEstimator est_pos[3];
CEstimator est_ypr[3];

int use_abs_pos = 1;
t_p5_vec3 cen, ypr, gpos, pos, zeropos, zerorot;
int finger_bend[5];

void ProcessGloveInput()
{
  int k;
  for (k=0;k<3;k++)
  {
    est_pos[k].AddSample(gpos.v[k]);
    est_ypr[k].AddSample(ypr.v[k]);
  };

  pos.x = -(est_pos[0].GetEstimate() - zeropos.x);
  pos.y = +(est_pos[1].GetEstimate() - zeropos.y);
  pos.z = +(est_pos[2].GetEstimate() - zeropos.z);

  ypr.yaw = est_ypr[0].GetEstimate() - zerorot.yaw;
  ypr.pitch = est_ypr[1].GetEstimate() - zerorot.pitch;
  ypr.roll = -(est_ypr[2].GetEstimate() - zerorot.roll);
}

void GetGloveInput()
{
  cli_GetP5Data (0, &p5data);
  if (use_abs_pos)
  {
    gpos = p5data.m_absolute_pos;
    ypr = p5data.m_absolute_rot;
  }
  else
  {
    gpos = p5data.m_relative_pos;
    ypr = p5data.m_relative_rot;
  }

  for (int i=0;i<5;i++)
     finger_bend[i] = p5data.m_bend[i];

  ProcessGloveInput();
}

void InitGlove()
{
	cli_Init ();
	est_pos[0].Init (POS_FILTER_SIZE);
	est_pos[1].Init (POS_FILTER_SIZE);
	est_pos[2].Init (POS_FILTER_SIZE);
	est_ypr[0].Init (ROT_FILTER_SIZE);
	est_ypr[1].Init (ROT_FILTER_SIZE);
	est_ypr[2].Init (ROT_FILTER_SIZE);
}

void InitPRFilters ()
{
  int k;
  for (k=0;k<3;k++)
  {
    est_pos[k].ZeroAll();
    est_ypr[k].ZeroAll();
  };
}

void GetZeroPosition ()
{
  if (use_abs_pos)
  {
    zeropos = p5data.m_absolute_pos;
    zerorot.yaw = 0.0f;
    zerorot.pitch = 0.0f;
    zerorot.roll = 0.0f;
  }
  else
  {
    zeropos = p5data.m_relative_pos;
    zerorot = p5data.m_relative_rot;
  }
}

enum ObjectType { STAR, CROSSBAR, TARGET, AMMO, ANIMATION };

class Object3D
{
public:
    float x, y, z;
    float dx, dy, dz;
    float r, g, b;
    float size;
    float dsize;
    float depth;
    float rx, ry, rz;
    float drx, dry, drz;
    ObjectType type;
    int subtype;
    int nloops;
    
    int visible;
    void Draw();
};

float distance(const Object3D &a, const Object3D &b)
{
	return sqrt( (a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z) );
}

void Object3D::Draw()
{
	if (!visible) return;
	glPushMatrix();
	switch (type)
	{
		case CROSSBAR:
			glTranslatef(x, y, 0);
			glRotatef(rz, 0, 0, 1);
			glColor3f(r, g, b);
			glBegin(GL_LINES);
			glVertex3f(-size, 0, -depth);
			glVertex3f(size, 0, -depth);
			glVertex3f(0, -size/2.0, -depth);
			glVertex3f(0, +size/2.0, -depth);
			glEnd();

		break;
		case STAR:
			glBegin(GL_LINE_STRIP);
			glColor3f(r, g, b);
			glVertex3f(x, y, z);
			glColor3f(0, 0, 0);
			glVertex3f(x, y, z - 10.0f*dz);
			glEnd();
		break;
		case TARGET:
			glColor3f(r, g, b);
			glTranslatef(x, y, z);
			glRotatef(rx, 1, 0, 0);
			glRotatef(ry, 0, 1, 0);
			glRotatef(rz, 0, 0, 1);
			glutSolidSphere(size, 15, 15);
		break;
		case AMMO:
			glColor3f(r, g, b);
			glTranslatef(x, y, z);
			glutWireSphere(size, subtype+3, subtype+2);
		break;
		case ANIMATION:
			glColor3f(r, g, b);
			glTranslatef(x, y, z);
			glutWireSphere(size, 17, 17);
		break;
	}
	glPopMatrix();
}

#define NUM_STARS 500
#define NUM_TARGETS 30
#define MAX_AMMO 3
#define FRAMES_BETWEEN_AMMO 20

#define MAX_ANIMATIONS (MAX_AMMO*5)

float minimum_speed = 0.001;
float star_speed = 0.005;
float target_speed = 0.01;
float ammo_speed = 0.01;
float target_spread = 0.08;
float anim_max_size = 0.05;

Object3D s[NUM_STARS];
Object3D t[NUM_TARGETS];
Object3D cross;
Object3D ammo[5][MAX_AMMO];
Object3D anim[MAX_ANIMATIONS];
int b_shooting[5];
int w=500, h=500;
int num_targets_hit = 0;
int is_fullscreen = 0;

float GetRandom(float low, float high)
{
	return low + (high - low) * (float)rand() / (float)(RAND_MAX);
}

void renderBitmapString (float x, float y, float z, void *font, char *s) 
{
	char c;
	glRasterPos3f(x, y, z);
	for (c=0; s[c] != 0; c++) 
	{
		glutBitmapCharacter(font, s[c]);
  	}
}

void CreateTarget(int i)
{
		t[i].x = GetRandom(-target_spread,target_spread);
		t[i].y = GetRandom(-target_spread,target_spread);
		t[i].z = GetRandom(-0.5,0.5);

		t[i].dx = 0;
		t[i].dy = 0;
		t[i].dz = GetRandom(minimum_speed,target_speed);

		t[i].r = GetRandom(0,1);
		t[i].g = GetRandom(0,1);
		t[i].b = GetRandom(0,1);
		
		t[i].rx = 0;
		t[i].ry = 0;
		t[i].rz = 0;

		t[i].drx = GetRandom(-10,10);
		t[i].dry = GetRandom(-10,10);
		t[i].drz = GetRandom(-10,10);
		
		t[i].size = GetRandom(0,0.01);
		
		t[i].type = TARGET;
		t[i].visible = 1;
		t[i].nloops = 0;
}

void Init3DData() 
{
	unsigned int i,j;

	cross.x = 0;
	cross.y = 0;
	cross.z = 0;
	cross.r = 0.8;
	cross.g = 0.8;
	cross.b = 0.8;
	cross.size = .05;
	cross.depth = .4;
	cross.type = CROSSBAR;
	cross.visible = 1;

	for(i = 0; i < MAX_ANIMATIONS; i++)
	{
		anim[i].type = ANIMATION;
		anim[i].dsize = 0.001;
		anim[i].visible = 0;
	}

	for(i = 0; i < 5; i++)
	for(j = 0; j < MAX_AMMO; j++)
	{
		ammo[i][j].x=0;
		ammo[i][j].y=0;
		ammo[i][j].z=0;
		ammo[i][j].r = (float)i/5.0;
		ammo[i][j].g = (float)i/5.0;
		ammo[i][j].b = (float)(5.0-i)/5.0;
		ammo[i][j].dz = ammo_speed;
		ammo[i][j].size = .003;
		ammo[i][j].type = AMMO;
		ammo[i][j].subtype = i;
		ammo[i][j].visible = 0;
		b_shooting[i] = 0;
	}

	for(i = 0; i < NUM_STARS; i++) 
	{
		s[i].x = GetRandom(-0.5,0.5);
		s[i].y = GetRandom(-0.5,0.5);
		s[i].z = GetRandom(-0.5,0.5);

		s[i].dx = 0;
		s[i].dy = 0;
		s[i].dz = GetRandom(minimum_speed,star_speed);

		s[i].g = 0;
		s[i].b = 300.0f*s[i].dz;
		s[i].r = s[i].b / 3.0f;
		
		s[i].type = STAR;
		s[i].visible = 1;
	}
	
	for(i = 0; i < NUM_TARGETS; i++) 
	{
		CreateTarget(i);
	}
}

void reshape(int width, int height) {
	w = width;
	h = height;
	glClearColor(0, 0, 0, 0);
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1, 1,-1, 1,-1, 1000);
	glFrustum(-1, 1,-1, 1, .7, 20);
	glMatrixMode(GL_MODELVIEW);
	//glLoadIdentity();
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_DEPTH_TEST);
}

void init() 
{
	InitGlove();
	reshape(w,h);
	Init3DData();
	GetGloveInput();
	GetZeroPosition();
}

void ProcessMotion()
{
	unsigned int i,j;
	cross.x = pos.x / 300.0f;
	cross.y = pos.y / 300.0f;
	cross.rz = ypr.roll;
	
	for(i = 0; i < 5; i++)
	if ( ( finger_bend[i] > 15) && (b_shooting[i] == 0) )
	{
		b_shooting[i] = 1;
		for(j = 1; j < MAX_AMMO; j++)
			ammo[i][j] = ammo[i][j-1];
		ammo[i][0].z = -0.05;
		ammo[i][0].visible = 1;
		float radius = ( (float)i-3.0f) / 100.0f;
		ammo[i][0].x = cross.x/2.0f + radius*cos(ypr.roll*3.14f/180.0f);
		ammo[i][0].y = cross.y/2.0f + radius*sin(ypr.roll*3.14f/180.0f);
	}
}

void DisplayResults()
{
	char buffer[1000];
	sprintf(buffer,"%d targets hit",num_targets_hit);
	glColor3f(0, 1, 1);
	renderBitmapString(-0.9,0.90,-0.7,GLUT_BITMAP_8_BY_13, buffer);
	//renderBitmapString(-0.9,0.85,-0.7,GLUT_BITMAP_8_BY_13, "L2");
}

void display(void)
{
	GetGloveInput();
	ProcessMotion();
	unsigned int i,j;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);

	DisplayResults();

	cross.Draw();
	for (i = 0; i < NUM_STARS; i++) 
		s[i].Draw();
	for (i = 0; i < NUM_TARGETS; i++) 
		t[i].Draw();
	for(i = 0; i < 5; i++)
	for(j = 0; j < MAX_AMMO; j++)
		ammo[i][j].Draw();

	for (i = 0; i < MAX_ANIMATIONS; i++) 
		anim[i].Draw();


	// check ammo
	for(i = 0; i < 5; i++)
	if (b_shooting[i] > 0)
	{
		b_shooting[i] ++;
		if (b_shooting[i] == FRAMES_BETWEEN_AMMO)
			b_shooting[i] = 0;
	}

	
	glutSwapBuffers();
}

void MoveStars()
{
	unsigned int i; 

	// star animation
	for(i = 0; i < NUM_STARS; i++) 
	{
		s[i].z += s[i].dz;
		s[i].z = s[i].z > 0 ? -0.5f : s[i].z;
	}
}

void MoveTargets()
{
	unsigned int i; 

	// big targets
	for(i = 0; i < NUM_TARGETS; i++) 
	{
		t[i].z += t[i].dz;
		t[i].rx += t[i].drx;
		t[i].ry += t[i].dry;
		t[i].rz += t[i].drz;
		
		//t[i].z = t[i].z > 0.5 ? -0.5 : t[i].z;
		if (t[i].z > 0.5)
		{
			t[i].z = -0.5; // loop again
			t[i].nloops ++;
			
			// if it spent too many cycles, regenerate it
			if (t[i].nloops > 5)
				CreateTarget(i);
		}
	}
}
void MoveAmmo()
{
	unsigned int i,j; 
	// ammo
	for(i = 0; i < 5; i++)
	for(j = 0; j < MAX_AMMO; j++)
	{
		ammo[i][j].z -= ammo[i][j].dz;
		if (ammo[i][j].z < -0.9f)
		{
			ammo[i][j].visible=0;
			b_shooting[i] = 0;
		}
	}
}

void AnimateAnimations()
{
	unsigned int i; 

	// collision animations
	for(i = 0; i < MAX_ANIMATIONS; i++)
	{
		if (anim[i].visible)
		{
			anim[i].size += anim[i].dsize;
			if (anim[i].size > anim_max_size)
				anim[i].visible = 0;
		}
		
	}
}

void Collide( int k, int j, int i)
{
	// collide ammo[k][j] with t[i]
	
	num_targets_hit ++;
	unsigned int l;
	ammo[k][j].visible = 0;
	CreateTarget(i);
	for(l = 0; l < MAX_ANIMATIONS; l++) 
	{
		if (anim[l].visible == 0)
			break;
	}
	if (l < MAX_ANIMATIONS) // we have at least a free slot
	{
		anim[l].visible = 1;
		anim[l].x = (ammo[k][j].x + t[i].x ) / 2.0;
		anim[l].y = (ammo[k][j].y + t[i].y ) / 2.0;
		anim[l].z = (ammo[k][j].z + t[i].z ) / 2.0;
		anim[l].r = t[i].r;
		anim[l].g = t[i].g;
		anim[l].b = t[i].b;
		anim[l].size = t[i].size;
	}
}

void ProcessColisions()
{
	unsigned int i,j,k; 

	// check for collisions
	for(i = 0; i < NUM_TARGETS; i++) 
	for(j = 0; j < MAX_AMMO; j++)
	for(k = 0; k < 5; k++)
	if (t[i].visible && ammo[k][j].visible)
	{
		if ( distance(ammo[k][j],t[i]) < ammo[k][j].size + t[i].size )
		{
			Collide(k, j, i);
		}
	}
}

void idle(void) 
{
	MoveStars();
	MoveTargets();
	MoveAmmo();
	AnimateAnimations();
	ProcessColisions();

	glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
	{
		case 27: 
			exit(0);
			break;
		case 'a':
			use_abs_pos = 1;
			GetZeroPosition ();
			InitPRFilters();
			GetGloveInput();
			break;
		case 'r':
			use_abs_pos = 0;
			GetZeroPosition ();
			InitPRFilters();
			GetGloveInput();
			break;
	}
}

int main(int argc, char** argv)
{
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(w, h);
	glutInit(&argc, argv);

	glutCreateWindow("P5 Starfield Demo");
	glutDisplayFunc(display);
	glutIdleFunc(idle);

	glutKeyboardFunc(keyboard);
	glutReshapeFunc(reshape);

	init();
	glutMainLoop();
	cli_Close ();
	return 0;
}
